/***************************************************************************
                          cfdb.c  -  description
                             -------------------
    begin                : Sat Jul  1 09:18:20 EST 2000
    copyright            : (C) 2000 by Ashley Cambrell
    email                : bec_ash@bigpond.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <libpq-fe.h>
#include "cfdb.h"

enum {
  CFDB_MOVED,
  LAST_SIGNAL
};

enum {
	MOVE_NONE,
	MOVE_START,
	MOVE_PREV,
	MOVE_NEXT,
	MOVE_LAST,
	MOVE_SPECIFIC
};

enum {
	RELOAD_CONTINUE,
	RELOAD_ALL,
	RELOAD_ONLY
};


enum {
	ARG_0,
	ARG_PRIMARYKEY,
	ARG_SQL_STRING,
	ARG_USER,
	ARG_CURSOR_NAME,
	ARG_TABLE
};

enum {
	MODE_NORMAL,
	MODE_INSERT,
	MODE_UPDATE
};



static void cfdb_class_init     			(CfDbClass *klass);
static void cfdb_init           			(CfDb      *cf);
static void cfdb_set_arg							(GtkObject *object, GtkArg *arg, guint arg_id);
static void cfdb_get_arg							(GtkObject *object, GtkArg *arg, guint arg_id);

static void cfdb_set_parent_cfdb 			(GtkWidget *cf, GtkWidget* parent);
static GtkWidget* cfdb_get_parent_cfdb			(GtkWidget *cf);

static void cfdb_record_start					(GtkWidget *widget, CfDb *cf);
static void cfdb_record_prev					(GtkWidget *widget, CfDb *cf);
static void cfdb_record_next					(GtkWidget *widget, CfDb *cf);
static void cfdb_record_last					(GtkWidget *widget, CfDb *cf);

static void cfdb_change_count_specific(CfDb *cf, int moveDirection, int position);
static void cfdb_change_count					(CfDb *cf, int moveDirection);

static void cfdb_create_cursor				(GtkWidget *cf);
static void cfdb_create_sql_string		(GtkWidget *cf);
static void cfdb_create_update_string	(GtkWidget *cf, GtkWidget *changed);
static gchar* cfdb_create_lookup_criteria(GtkWidget *cf);
static gchar* cfdb_create_persistant_lookup_criteria(GtkWidget *cf);

static void cfdb_do_reload(GtkWidget *cf, int reload_type);
static void cfdb_reload_child(GtkWidget *parent_cfdb, GtkWidget *child_cfdb);

static void cfdb_block_all_widgets		(GtkWidget *cf);
static void cfdb_unblock_all_widgets	(GtkWidget *cf);

static void cfdb_widget_changed				(GtkWidget *changed, GtkWidget *cf);
static void cfdb_widget_lost_focus		(GtkWidget *changed, GdkEventFocus *event, gpointer acf);
static void cfdb_widgets_activate			(GtkWidget *widget, GtkWidget *cf);

static void cfdb_hash_insert_string		(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_clear_all_widgets(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_free						(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_widget_block		(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_widget_unblock	(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_field_lookup		(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_get_registered_fields(gpointer key, gpointer value, gpointer user_data);
static void cfdb_hash_count_registered_fields(gpointer key, gpointer value, gpointer user_data);

static void cfdb_slist_free						 (gpointer data, gpointer user_data);

static gint cfdb_signals[LAST_SIGNAL] = { 0 };

/********************
*
*  public functions
*
*********************/

guint
cfdb_get_type ()
{
  static guint cfdb_type = 0;

  if (!cfdb_type)
  {
  	GtkTypeInfo cfdb_info =
  	{
			"CfDb",
			sizeof (CfDb),
			sizeof (CfDbClass),
			(GtkClassInitFunc) cfdb_class_init,
			(GtkObjectInitFunc) cfdb_init,
      (GtkArgSetFunc) NULL,
      (GtkArgGetFunc) NULL
    };

      cfdb_type = gtk_type_unique (gtk_vbox_get_type (), &cfdb_info);
	}

  return cfdb_type;
}

/* old cfdb_new.  Changed so we could easily use widget with glade (one initializer) */
//GtkWidget*
//cfdb_new ()
//{
//  return GTK_WIDGET ( gtk_type_new (cfdb_get_type ()));
//}

GtkWidget*
cfdb_new (gchar *conn_info, gchar *primarykey, gchar *table) ///////, GtkWidget *parent_cfdb)
{
  CfDb *cf;
  gchar **tmpstring;
  int havematch = 0, i;

 	g_return_val_if_fail(conn_info != NULL, NULL);
 	g_return_val_if_fail(primarykey != NULL, NULL);
 	g_return_val_if_fail(table != NULL, NULL);
	
	
  cf = gtk_type_new (cfdb_get_type ());

  tmpstring = g_strsplit(conn_info, " ", 9);
	
  /* check to see if the user has been set in the conn info */
  for ( i = 0 ; i < 10 && tmpstring[i] != NULL ; i++ )
  {
  	int match = g_strncasecmp(tmpstring[i],"user",4);
  	if ( match == 0 )
  	{
  		/* if there is a match */
  		gchar **matchstring;

			/* split it again based on = so we can separate the actual username */
  		matchstring = g_strsplit(tmpstring[i],"=",2);

  		/* double check and then set the cfdb user to this user */
  		if ( g_strcasecmp(matchstring[0],"user") ==	0 )
  		{
  			/* use g_strdup so we can free the memory properly */
  			cf->user = g_strdup(matchstring[1]);
  			havematch++;
  		}
 		  g_strfreev(matchstring);
  	}
  }
  g_strfreev(tmpstring);

  /* if the user wasn't set then we take the user variable in cfdb and add it to the conn_info string */
  if ( havematch == 0 )
  {
  	if ( cf->user == NULL )
  	{
			cf->user = getenv("USER");
		}
		conn_info = g_strconcat(conn_info, g_strdup_printf(" %s", cf->user));
	}
	cf->conn_info = conn_info;
  cf->conn = PQconnectdb(cf->conn_info);
  if (PQstatus(cf->conn) == CONNECTION_BAD)
	{
		g_warning("Connection to database '%s' failed. Connection string: %s\n", PQdb(cf->conn), cf->conn_info);
		g_warning("%s", PQerrorMessage(cf->conn));
		PQfinish(cf->conn);
		return NULL;
	}
	
	cf->primarykey = primarykey;
	
	cf->table = table;
	
//////	if ( parent_cfdb != NULL )
//////	{
//////		if ( IS_CFDB(parent_cfdb) )
//////		{
//////			cf->parent_cfdb = parent_cfdb;
//////		}
//////		else
//////		{
//////			g_warning("Parent not CFDB. Parent not set\n");
//////		}
//////  }

	return GTK_WIDGET(cf);
}

static void	
cfdb_set_parent_cfdb (GtkWidget *cf, GtkWidget* parent)
{
	CFDB(cf)->parent_cfdb = parent;
}

static GtkWidget*	
cfdb_get_parent_cfdb (GtkWidget *cf)
{
	return CFDB(cf)->parent_cfdb;
}

void
cfdb_set_primarykey(GtkWidget *cf, gchar *primarykey)
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	CFDB(cf)->primarykey = primarykey;
}

void
cfdb_set_primaryley_value(GtkWidget* cf, gchar* primarykey_value)
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(primarykey_value != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	CFDB(cf)->primarykey_value = primarykey_value;	
}

static void
cfdb_reload_child(GtkWidget *parent_cfdb, GtkWidget *child_cfdb)
{
	cfdb_do_reload(child_cfdb, RELOAD_ALL);
}

void
cfdb_reload(GtkWidget *cf)
{
	cfdb_do_reload(cf, RELOAD_ALL);
}

void
cfdb_reload_continue(GtkWidget *cf)
{
	cfdb_do_reload(cf, RELOAD_CONTINUE);
}

void
cfdb_start(GtkWidget *cf)
{
 cfdb_do_reload(cf, RELOAD_ALL);
}

void
cfdb_set_conn(GtkWidget *cf, PGconn *conn)
{
	if (PQstatus(conn) == CONNECTION_BAD)
	{
		g_warning("Bad connection!!\n");
		g_warning("%s", PQerrorMessage(conn));
	}
	else
	{
		CFDB(cf)->conn = conn;
	}
}

void
cfdb_add_field_widget_map (GtkWidget *cf, gchar* field, GtkWidget* widget)
{
	/* insert into field widget hash map */
	g_return_if_fail(cf != NULL);
	g_return_if_fail(widget != NULL);
	g_return_if_fail(field != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	g_hash_table_insert(CFDB(cf)->field_widget_hash, field, widget);
	
	/* insert into widget field hash map */
  g_hash_table_insert(CFDB(cf)->widget_field_hash, widget, field);

  if ( GTK_IS_EDITABLE(widget) ) /* if widget is of editable variety */
  {
  	/* catch changed event so we know if we have to do an create an update_string when we leave focus */
  	gtk_signal_connect(GTK_OBJECT(widget), "changed",  GTK_SIGNAL_FUNC(cfdb_widget_changed), cf);
  	
  	/* catch focus_out_event so we know if we have to create an update_string */
 		gtk_signal_connect(GTK_OBJECT(widget), "focus_out_event",  GTK_SIGNAL_FUNC(cfdb_widget_lost_focus),(gpointer)cf);

 		/* catch the activate event so we know if we have to do an insert if we are in MODE_INSERT */ 		
 		gtk_signal_connect(GTK_OBJECT(widget), "activate",  GTK_SIGNAL_FUNC(cfdb_widgets_activate),(gpointer)cf);
  }
  else if ( IS_CFDB(widget) )
  {
		/* try and stop recursive situations */
		if ( widget != cf )
		{
				/* set the parent of the widget to cf */
			cfdb_set_parent_cfdb(widget,cf);
			gtk_signal_connect(GTK_OBJECT(cf), "moved",  GTK_SIGNAL_FUNC(cfdb_reload_child), (gpointer)widget);
		}
		else
			g_warning("widget == cf in cfdb_add_field_widget_map \n");

  }
  else
  {
  	g_warning("Dont know how to handle widget type\n");
  }
}

void
cfdb_set_table(GtkWidget *cf, gchar* table)
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));

	CFDB(cf)->table = table;
}

/********************
*
*  private functions
*
*********************/

static void
cfdb_class_init (CfDbClass *klass)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) klass;

  cfdb_signals[CFDB_MOVED] = gtk_signal_new ("moved",
					 GTK_RUN_FIRST,
					 object_class->type,
					 GTK_SIGNAL_OFFSET (CfDbClass,moved),
					 gtk_signal_default_marshaller,
      			GTK_TYPE_NONE, 0);


  gtk_object_class_add_signals (object_class, cfdb_signals, LAST_SIGNAL);
	
	/*setup args */
	gtk_object_add_arg_type ("CfDb::primarykey", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_PRIMARYKEY);
	gtk_object_add_arg_type ("CfDb::sql_string", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_SQL_STRING);
	gtk_object_add_arg_type ("CfDb::user", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_USER);
	gtk_object_add_arg_type ("CfDb::cursor_name", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_CURSOR_NAME);
	gtk_object_add_arg_type ("CfDb::table", GTK_TYPE_STRING, GTK_ARG_READWRITE, ARG_TABLE);

	object_class->set_arg = cfdb_set_arg;
	object_class->get_arg = cfdb_get_arg;

  klass->moved = NULL;
}

static void
cfdb_init (CfDb *cf)
{
	GtkWidget *table;
 	GtkWidget *frame;

 	/* setup visuals */
 	frame = gtk_frame_new(NULL);
 	gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_ETCHED_IN);
 	gtk_container_add(GTK_CONTAINER(cf),frame);
 	gtk_widget_show(frame);
	table = gtk_table_new (3, 5, FALSE);
	gtk_container_add (GTK_CONTAINER(frame), table);
	gtk_widget_show (table);
	
	cf->buttonStart = gtk_button_new_with_label("|<-");
	cf->buttonPrev = gtk_button_new_with_label("<-");
	cf->buttonNext = gtk_button_new_with_label("->");
	cf->buttonLast = gtk_button_new_with_label("->|");
	cf->labelRecordCounter = gtk_label_new(NULL);

//	gtk_widget_set_usize(cf->buttonStart,10,10);
//	gtk_widget_set_usize(cf->buttonPrev,10,10);
//	gtk_widget_set_usize(cf->buttonNext,10,10);
//	gtk_widget_set_usize(cf->buttonLast,10,10);		
	gtk_table_attach(GTK_TABLE(table),cf->buttonStart,0,1,1,2,GTK_FILL,GTK_FILL,1,1);
	gtk_table_attach(GTK_TABLE(table),cf->buttonPrev,1,2,1,2,GTK_FILL,GTK_FILL,1,1);
	gtk_table_attach(GTK_TABLE(table),cf->buttonNext,3,4,1,2,GTK_FILL,GTK_FILL,1,1);
	gtk_table_attach(GTK_TABLE(table),cf->buttonLast,4,5,1,2,GTK_FILL,GTK_FILL,1,1);
	gtk_table_attach(GTK_TABLE(table),cf->labelRecordCounter,2,3,1,2,GTK_FILL|GTK_EXPAND|GTK_SHRINK,GTK_FILL|GTK_EXPAND|GTK_SHRINK,1,1);

	gtk_widget_show(cf->buttonStart);
	gtk_widget_show(cf->buttonPrev);
	gtk_widget_show(cf->buttonNext);
	gtk_widget_show(cf->buttonLast);
	gtk_widget_show(cf->labelRecordCounter);

	/* connect buttons to record movers */	
	gtk_signal_connect(GTK_OBJECT(cf->buttonStart), "clicked", GTK_SIGNAL_FUNC(cfdb_record_start), cf);
	gtk_signal_connect(GTK_OBJECT(cf->buttonPrev), "clicked", GTK_SIGNAL_FUNC(cfdb_record_prev), cf);
	gtk_signal_connect(GTK_OBJECT(cf->buttonNext), "clicked", GTK_SIGNAL_FUNC(cfdb_record_next), cf);
	gtk_signal_connect(GTK_OBJECT(cf->buttonLast), "clicked", GTK_SIGNAL_FUNC(cfdb_record_last), cf);
		
	/* setup default cfdb values */
	cf->totalCount = 100;
	cf->currentCount = 1;
	cf->primarykey = "\0";
	cf->parent_cfdb = NULL;
	cf->cursor_name = NULL;
	cf->user = getenv("USER");
	cf->table = NULL;
	cf->res = NULL;
	cf->field_widget_hash = g_hash_table_new(g_str_hash, g_str_equal);
	cf->widget_field_hash  = g_hash_table_new(g_direct_hash, g_direct_equal);
	cf->being_reloaded = FALSE;
//	cf->changed_field_values = NULL;
	cf->field_changed = FALSE;
	cf->current_mode = MODE_NORMAL;	
	cf->insert_string1 = NULL;
	cf->insert_string2 = NULL;
	cf->insert_field_value_hash  = g_hash_table_new(g_str_hash, g_str_equal);
	cf->lookup_criteria = NULL;
	cf->conn_info = NULL;
	cf->persistant_lookup_criteria = NULL;
}


static void
cfdb_set_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
	CfDb* cf = NULL;
	
	g_return_if_fail(cf != NULL);
	g_return_if_fail(arg != NULL);
	g_return_if_fail(IS_CFDB(cf));
 	
	cf = CFDB(object);
 	
 	switch( arg_id )
 	{
 		case ARG_PRIMARYKEY:
 			cf->primarykey = GTK_VALUE_STRING(*arg);
 			break;
 		case ARG_SQL_STRING:
 			cf->sql_string = GTK_VALUE_STRING(*arg);
 			break;
 		case ARG_USER:
 			cf->user = GTK_VALUE_STRING(*arg);
 			break;
 		case ARG_CURSOR_NAME:
 			cf->cursor_name = GTK_VALUE_STRING(*arg);
 			break;
 		case ARG_TABLE:
 			cf->table = GTK_VALUE_STRING(*arg);
 			break;
		default:
 			break;
	}	

} 	

static void
cfdb_get_arg(GtkObject *object, GtkArg *arg, guint arg_id)
{
 	CfDb* cf = NULL;
	
	g_return_if_fail(cf != NULL);
	g_return_if_fail(arg != NULL);
//	g_return_if_fail(IS_CFDB(cf));
 	
 	cf = CFDB(object);
 	
 	switch( arg_id )
 	{
 		case ARG_PRIMARYKEY:
 			GTK_VALUE_STRING(*arg) = cf->primarykey;
 			break;
 		case ARG_SQL_STRING:
 			GTK_VALUE_STRING(*arg) = cf->sql_string;
 			break;
 		case ARG_USER:
 			GTK_VALUE_STRING(*arg) = cf->user;
 			break;
 		case ARG_CURSOR_NAME:
 			GTK_VALUE_STRING(*arg) = cf->cursor_name;
 			break;
 		case ARG_TABLE:
 			GTK_VALUE_STRING(*arg) = cf->table;
 			break;
 		default:
 			arg->type = GTK_TYPE_INVALID;
 			break;
 	}

}

static void
cfdb_record_start(GtkWidget *widget, CfDb *cf)
{

	g_return_if_fail(cf != NULL);

	cfdb_change_count(cf, MOVE_START);
	if (cf->being_reloaded == FALSE)	
		gtk_signal_emit(GTK_OBJECT(cf), cfdb_signals[CFDB_MOVED]);

}

static void
cfdb_record_prev(GtkWidget *widget, CfDb *cf)
{
	g_return_if_fail(cf != NULL);
	
	cfdb_change_count(cf, MOVE_PREV);
	if (cf->being_reloaded == FALSE)	
		gtk_signal_emit(GTK_OBJECT(cf), cfdb_signals[CFDB_MOVED]);
}

static void
cfdb_record_next(GtkWidget *widget, CfDb *cf)
{
	g_return_if_fail(cf != NULL);
		
	cfdb_change_count(cf, MOVE_NEXT);
	if (cf->being_reloaded == FALSE)	
		gtk_signal_emit(GTK_OBJECT(cf), cfdb_signals[CFDB_MOVED]);
}

static void
cfdb_record_last(GtkWidget *widget, CfDb *cf)
{
	g_return_if_fail(cf != NULL);
		
	cfdb_change_count(cf, MOVE_LAST);
	if (cf->being_reloaded == FALSE)	
		gtk_signal_emit(GTK_OBJECT(cf), cfdb_signals[CFDB_MOVED]);
}

static void
cfdb_record_specific(GtkWidget *widget, CfDb *cf, int specific_record)
{
	g_return_if_fail(cf != NULL);
	
	cfdb_change_count_specific(cf, MOVE_SPECIFIC, specific_record);
	if (cf->being_reloaded == FALSE)	
		gtk_signal_emit(GTK_OBJECT(cf), cfdb_signals[CFDB_MOVED]);
}

static void
cfdb_change_count(CfDb *cf, int moveDirection)
{
	g_return_if_fail(cf != NULL);
	
	cfdb_change_count_specific(cf, moveDirection, -1);
}


static void
cfdb_change_count_specific(CfDb *cf, int moveDirection, int position )
{
	int move = 0;
	int fetch = 0;
  int nFields = 0;
//	char *tmpa = NULL;
	int i, j;
//	gchar *tmpstring = NULL;
	gboolean zero_records = FALSE;
	
	/* if there are fields that have changed value */
	if ( cf->number_fields_update > 0 && cf->update_string != NULL )
	{
		/* complete the update_string, already partitally created when an widget has changed then lost focus*/
		cf->update_string = g_strdup_printf("UPDATE %s %s WHERE %s = '%s'", cf->table, cf->update_string, cf->primarykey, cf->primarykey_value);

		/* reset the number of updates back to 0 */
    cf->number_fields_update = 0;
		
		/* execute the update string */
		cf->res = PQexec(cf->conn, cf->update_string);
		if (!cf->res || PQresultStatus(cf->res) != PGRES_COMMAND_OK)
		{
			g_warning("UPDATE command failed: %s\n", cf->update_string);
			g_warning("Error Message: %s\n",PQresultErrorMessage(cf->res));
			g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(cf->res)));
			PQclear(cf->res);
		}
		PQclear(cf->res);
		
		/* set the update string to NULL */
		g_free(cf->update_string);
		cf->update_string = NULL;
		
		/* do a reload only, so changes are committed */
		cfdb_do_reload(GTK_WIDGET(cf), RELOAD_CONTINUE);
		
		/* every thing should be OK to go any direction. */		
	}
	
	/* if we are going to the start of the records */
	if (moveDirection == MOVE_START)
	{
		/* move to record number 2*/
		move = -cf->currentCount;
		
		/* then fetch fetch back one more and you have record number 1 */
		fetch = +1;
	}
	else if( moveDirection == MOVE_PREV )
	{
		move = 0;
		
		fetch = -1;
	}
	else if( moveDirection == MOVE_NEXT )
	{
		move = 0;
		
		fetch = +1;
	}
	else if( moveDirection == MOVE_LAST )
	{
		move = cf->totalCount - cf->currentCount -1;
		
		fetch = +1;
	}
	else if( moveDirection == MOVE_SPECIFIC && position != -1 )
	{
		if (  position  > cf->totalCount )
		{
			position = cf->totalCount;
		}
		if ( position > 0 )
		{
			move = position - cf->currentCount;
			if ( move > 0 )
			{
				/* move is +ve so we take one away from move and make fetch +ve */
				move--;
				fetch = +1;
			}
			else if ( move < 0 )
			{
				/* move is negative so we add one and make fetch -ve */
				move++;
				fetch = -1;
			}
			else
			{
				move = position - 1;
				fetch = +1;
				cf->currentCount = 0;
			}
		}
		else
		{
			g_warning("MOVE_SPECIFIC: position is less than 1 or greater than totalCount: %d\n",position);
		}
	}
	
//	g_print("cc: %d  tc: %d  mv: %d  fe: %d\n",cf->currentCount, cf->totalCount, move, fetch);
	
	/* start MOVE and FETCH */
	if (((cf->currentCount + move + fetch) > 0) && ((cf->currentCount + move + fetch) <= cf->totalCount))
	{
		if ( move != 0 )
		{
			cf->res = PQexec(cf->conn, g_strdup_printf("MOVE %d in %s",move, cf->cursor_name));
			if (!cf->res || PQresultStatus(cf->res) != PGRES_COMMAND_OK)
			{
				g_warning("MOVE command wasn't successful\n");
				g_warning("Error Message: %s\n",PQresultErrorMessage(cf->res));
				g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(cf->res)));
				PQclear(cf->res);
			}
			PQclear(cf->res);
		}
		cf->res = PQexec(cf->conn, g_strdup_printf("FETCH %d in %s",fetch, cf->cursor_name));
		if (!cf->res || PQresultStatus(cf->res) != PGRES_TUPLES_OK)
		{
			g_warning("FETCH command didn't return tuples properly: sql: %s cursor: %s pk: %s pkv: %s\n", cf->sql_string, cf->cursor_name, cf->primarykey, cf->primarykey_value);
			g_warning("Error Message: %s\n",PQresultErrorMessage(cf->res));
			g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(cf->res)));
			PQclear(cf->res);
		}
		cf->currentCount += (move + fetch);		
	}
	
	if ( cf->totalCount == 0 )
	{
		zero_records = TRUE;
		cf->currentCount = 0;
  }

	gtk_label_set_text(GTK_LABEL(cf->labelRecordCounter), g_strdup_printf("%d of %d",cf->currentCount,cf->totalCount));
	if ( zero_records == FALSE )
	{
  	nFields = PQnfields(cf->res);

   	/* block signals so we don't get any "changed" signals emitted */
    cfdb_block_all_widgets(GTK_WIDGET(cf));
   	
   	/* update all the widgets with the new values */
	  for (i = 0; i < PQntuples(cf->res); i++)
		{
				for (j = 0; j < nFields; j++)
				{
					GtkWidget *tmpWidget = g_hash_table_lookup(cf->field_widget_hash, PQfname(cf->res, j));
					if ( GTK_IS_ENTRY(tmpWidget) )
					{
						/* set the new value of the GtkEntry widget */
						gtk_entry_set_text(GTK_ENTRY(tmpWidget),PQgetvalue(cf->res, i, j));
					}
					else if ( IS_CFDB(tmpWidget) )
					{
						/* set the new value of the primarykey and then reload the child */
						cfdb_set_primaryley_value(tmpWidget, PQgetvalue(cf->res, i, j));
						//cfdb_reload(GTK_WIDGET(cf),tmpWidget);
					}
					tmpWidget = NULL;
				}
		}
		cfdb_unblock_all_widgets(GTK_WIDGET(cf));
		cf->primarykey_value = PQgetvalue(cf->res, 0,PQfnumber(cf->res, cf->primarykey));
	}
}


static void
cfdb_do_reload(GtkWidget *cf, int reload_type)
{
	int nFields = 0;
	int i,  insert_record = -1;
	gchar* oid = NULL;
	gchar* sql_string = NULL;
	
	CFDB(cf)->being_reloaded = (gboolean*) TRUE;

//////	if( CFDB(cf)->parent_cfdb != NULL )
//////	{
//////		g_warning("Parent not NULL\n");
//////	}
//////	if ( CFDB(cf)->conn  == NULL )
//////	{
//////		g_warning("conn NULL!!\n");
//////	}
	
	if ( CFDB(cf)->current_mode == MODE_INSERT )
		g_print("CFDB(cf)->current_mode == MODE_INSERT\n");
	if ( CFDB(cf)->insert_string1 != NULL )
		g_print("CFDB(cf)->insert_string1 != NULL\n");	
	
	/* do a big check to see if we are in the right mode for an insert*/
	if ( CFDB(cf)->current_mode == MODE_INSERT && CFDB(cf)->insert_string1 != NULL )// && CFDB(cf)->insert_string1 == NULL)
	{
		CFDB(cf)->res = PQexec(CFDB(cf)->conn, CFDB(cf)->insert_string1);
		if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
		{
			g_warning("INSERT command failed: %s\n",CFDB(cf)->insert_string1);
			g_warning("Error Message: %s\n",PQresultErrorMessage(CFDB(cf)->res));
			g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(CFDB(cf)->res)));
			PQclear(CFDB(cf)->res);
	  }
	  g_free(CFDB(cf)->insert_string1);
	  CFDB(cf)->insert_string1 = NULL;
	  oid = g_strdup(PQoidStatus(CFDB(cf)->res));
		PQclear(CFDB(cf)->res);
	}

	if ( CFDB(cf)->cursor_name != NULL )
	{
		/* close cursor if cursor_name is defined */
		sql_string = g_strdup_printf("CLOSE %s",CFDB(cf)->cursor_name);
		CFDB(cf)->res = PQexec(CFDB(cf)->conn, sql_string);
		if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
		{
			g_warning("CLOSE command failed: %s \n", sql_string);
			PQclear(CFDB(cf)->res);
	  }	
		
		g_free(sql_string);
		
		PQclear(CFDB(cf)->res);
		
		/* Do commit after close */
		CFDB(cf)->res = PQexec(CFDB(cf)->conn, "COMMIT");
		if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
		{
			g_warning("COMMIT command failed. No BEGIN done yet? ie. first reload\n");
			PQclear(CFDB(cf)->res);
	  }	
	}
	
	/*************************************************************************
	*
	* Start Hack to get total record count (it is actually now used to get the
	* record number of a newly inserted field)
	*
	**************************************************************************/	
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, "BEGIN");
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("BEGIN command failed\n");
		PQclear(CFDB(cf)->res);
	}
	
	/* create a new unique cursor name (based on username and current time since epoch in secs) */
	cfdb_create_cursor(cf);
	

	/* create a new sql string based on the fieldnames registered with widgets */
	cfdb_create_sql_string(cf);
	
	/* declare the cursor */
	g_print("THE SQL: %s\n", CFDB(cf)->sql_string);
	sql_string = g_strdup_printf("DECLARE %s CURSOR FOR %s",CFDB(cf)->cursor_name,CFDB(cf)->sql_string);
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, sql_string);
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("DECLARE CURSOR command failed: %s\n", sql_string);
		g_warning("Error Message: %s\n",PQresultErrorMessage(CFDB(cf)->res));
		g_free(sql_string);
		g_error("Result Status: %s\n",PQresStatus(PQresultStatus(CFDB(cf)->res)));
		PQclear(CFDB(cf)->res);
	}
	g_free(sql_string);	
	PQclear(CFDB(cf)->res);
	
	/* fetch all the records so we can show the record count properley in the record count label */
  sql_string = g_strdup_printf("FETCH ALL in %s",CFDB(cf)->cursor_name);
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, sql_string);
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_TUPLES_OK)
	{
		g_warning("FETCH ALL command didn't return tuples properly: %s\n", sql_string);
		PQclear(CFDB(cf)->res);
	}
	g_free(sql_string);	
	
	/* update the record count label */
	CFDB(cf)->totalCount = PQntuples(CFDB(cf)->res);
	nFields = PQnfields(CFDB(cf)->res);
	if ( CFDB(cf)->current_mode == MODE_INSERT )
	{
		
		g_print("oid is field num: %d\n",PQfnumber(CFDB(cf)->res, "oid"));
		for (i = 0; i < PQntuples(CFDB(cf)->res) ; i++)
		{
			gchar *thisoid = PQgetvalue(CFDB(cf)->res, i, PQfnumber(CFDB(cf)->res, "oid"));
//			g_print("thisoid: %s oid: %s\n", thisoid,oid);
      if ( strcmp(thisoid, oid) == 0 )
	    {
	      insert_record = i + 1;
//	      g_print("record number: %d\n", insert_record);
	      break;
	    }
		}
		
		/* set the mode back to normal mode ready for normal operation */
		CFDB(cf)->current_mode = MODE_NORMAL;
	}/* end if MODE_INSERT */


	/* close this temporary cursor */
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, g_strdup_printf("CLOSE %s",CFDB(cf)->cursor_name));
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("CLOSE command failed. \n");
		PQclear(CFDB(cf)->res);
  }	
	PQclear(CFDB(cf)->res);
	
	/* do a COMMIT just to be safe */
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, "COMMIT");
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("CLOSE 2 command failed. No BEGIN done yet? ie. first reload\n");
		PQclear(CFDB(cf)->res);
  }	
	PQclear(CFDB(cf)->res);
	
	/*************************************************************************
	*
	* End Hack to get total record count
	*
	**************************************************************************/	

	CFDB(cf)->res = PQexec(CFDB(cf)->conn, "BEGIN");
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("BEGIN command failed\n");
		PQclear(CFDB(cf)->res);
	}
	PQclear(CFDB(cf)->res);
	
	sql_string = g_strdup_printf("DECLARE %s CURSOR FOR %s",CFDB(cf)->cursor_name, CFDB(cf)->sql_string);
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, sql_string);
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("DECLARE CURSOR command failed in reload: %s\n", sql_string);
		PQclear(CFDB(cf)->res);
	}
	PQclear(CFDB(cf)->res);
 	g_free(sql_string);
 	
	/* if insert_record is no -1 - do a load specific cause it means we have just done an insert and have found our position to go to
		ready for more action */
	if ( insert_record != -1 )
		cfdb_record_specific(NULL, CFDB(cf), insert_record);
	else if ( reload_type == RELOAD_ALL )
	 	cfdb_record_start(NULL, CFDB(cf));
	else if ( reload_type == RELOAD_CONTINUE )
		cfdb_record_specific(NULL, CFDB(cf), CFDB(cf)->currentCount);
	else if ( reload_type == RELOAD_ONLY )
	{
		/* do nothing */
		g_print("RELOAD_ONLY\n");
	}
	else
		g_print("Unknown reload type: %d\n",  reload_type);

	g_print("Near end of do_reload\n");
	/* set being_reloaded back to FALSE so we don't block future moved emits */
	CFDB(cf)->being_reloaded = FALSE;
	gtk_signal_emit(GTK_OBJECT(CFDB(cf)), cfdb_signals[CFDB_MOVED]);
	g_free(oid);
	g_print("At end of do_reload\n");
}

static void
cfdb_create_cursor(GtkWidget *cf)
{
	CfDb *tmpcf = CFDB(cf);
	time_t *t = NULL;
	t = g_malloc(sizeof(time_t));
	time(t);
	g_print("cursor: user is %s\n", tmpcf->user);
	tmpcf->cursor_name = g_strdup_printf("%s%d",tmpcf->user, (int)t);
	g_print("cursor_name: %s, %s\n",tmpcf->cursor_name, tmpcf->primarykey);
	g_free(t);
}

static void
cfdb_create_sql_string(GtkWidget *cf)
{
	GtkWidget *parent = NULL;
	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_field_lookup;

	/* if there is a parent, create an sql string that binds to the parents primary key.  This one work properley if the childs
			primary key isn't the same as the parents... should each cfdb hold the tables pk and the parents binding pk? */
	if ( CFDB(cf)->parent_cfdb != NULL && CFDB(cf)->primarykey_value != NULL )
	{
		g_free(CFDB(cf)->sql_string);
		CFDB(cf)->sql_string = NULL;
		g_hash_table_foreach(CFDB(cf)->field_widget_hash, hash_func, CFDB(cf));
		parent = cfdb_get_parent_cfdb(cf);
		if (CFDB(cf)->lookup_criteria != NULL || CFDB(cf)->persistant_lookup_criteria != NULL)
		{
			CFDB(cf)->sql_string = g_strdup_printf("SELECT oid, %s FROM %s WHERE (%s AND %s = '%s') ORDER BY %s",g_strdup_printf("%s, %s",CFDB(cf)->primarykey,CFDB(cf)->sql_string), CFDB(cf)->table, cfdb_create_lookup_criteria(cf),CFDB(parent)->primarykey, CFDB(parent)->primarykey_value, CFDB(cf)->primarykey);
		}
		else
		{
			CFDB(cf)->sql_string = g_strdup_printf("SELECT oid, %s FROM %s WHERE %s = '%s' ORDER BY %s",g_strdup_printf("%s, %s",CFDB(cf)->primarykey,CFDB(cf)->sql_string), CFDB(cf)->table,CFDB(parent)->primarykey, CFDB(parent)->primarykey_value, CFDB(cf)->primarykey);
		}
		g_print("Cfdb is child: %s\n",CFDB(cf)->sql_string);
		parent = NULL;
	}
	else
	{
		g_free(CFDB(cf)->sql_string);
		CFDB(cf)->sql_string = NULL;
		g_hash_table_foreach(CFDB(cf)->field_widget_hash, hash_func, CFDB(cf));
		if (CFDB(cf)->lookup_criteria != NULL || CFDB(cf)->persistant_lookup_criteria != NULL)
		{
			CFDB(cf)->sql_string = g_strdup_printf("SELECT oid, %s FROM %s WHERE %s ORDER BY %s",g_strdup_printf("%s, %s",CFDB(cf)->primarykey,CFDB(cf)->sql_string), CFDB(cf)->table, cfdb_create_lookup_criteria(cf), CFDB(cf)->primarykey);
		}
		else
		{
			CFDB(cf)->sql_string = g_strdup_printf("SELECT oid, %s FROM %s ORDER BY %s",g_strdup_printf("%s, %s",CFDB(cf)->primarykey,CFDB(cf)->sql_string), CFDB(cf)->table, CFDB(cf)->primarykey);
		}
	}
}

static gchar*
cfdb_create_lookup_criteria(GtkWidget *cf)
{
	GSList *tmp = NULL;
	int counter = 0;
	gchar *string = NULL;
	
	g_return_val_if_fail(cf != NULL,NULL);
	g_return_val_if_fail(IS_CFDB(cf),NULL);
	
	tmp = CFDB(cf)->lookup_criteria;
	
	while ( tmp != NULL )
	{
		CfDbSearch *criteria = tmp->data;
		g_print("Crteria field %s, value %s\n",criteria->field, criteria->value);
		
		if ( counter == 0 )
		{
			string = g_strdup_printf("(%s ~* '%s') ",criteria->field, criteria->value);
		}
		else
		{
			if ( criteria->boolval  == LOOKUP_AND )
			{
				string = g_strdup_printf("(%s AND %s ~* '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_OR )
			{
				string = g_strdup_printf("(%s OR %s ~* '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_NAND )
			{
				string = g_strdup_printf("(%s AND %s !~* '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_NOR )
			{
				string = g_strdup_printf("(%s OR %s !~* '%s') ",string, criteria->field, criteria->value);
			}
		}
		counter++;	
		tmp = g_slist_next(tmp);
	}
	/* check to see if there are any persistant criteria and add them at the end so they override all
		other criteria */
	if ( CFDB(cf)-> persistant_lookup_criteria != NULL )
	{
		if ( string != NULL )
		{
			/* just a straight AND b/c we want it to override everything */
			string = g_strdup_printf("%s AND %s ",string, cfdb_create_persistant_lookup_criteria(cf));
		}
		else
		{
			string = g_strdup_printf("%s ",cfdb_create_persistant_lookup_criteria(cf));
		}
  }

	g_print("lookup_criteria: %s\n",string);
	return string;
}

static gchar*
cfdb_create_persistant_lookup_criteria(GtkWidget *cf)
{
	GSList *tmp = NULL;
	int counter = 0;
	gchar *string = NULL;
	
	g_return_val_if_fail(cf != NULL,NULL);
	g_return_val_if_fail(IS_CFDB(cf),NULL);
	
	tmp = CFDB(cf)->persistant_lookup_criteria;
	
	while ( tmp != NULL )
	{
		CfDbSearch *criteria = tmp->data;
		g_print("Crteria field %s, value %s\n",criteria->field, criteria->value);
		
		if ( counter == 0 )
		{
			string = g_strdup_printf("(%s = '%s') ",criteria->field, criteria->value);
		}
		else
		{
			if ( criteria->boolval  == LOOKUP_AND )
			{
				string = g_strdup_printf("(%s AND %s = '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_OR )
			{
				string = g_strdup_printf("(%s OR %s = '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_NAND )
			{
				string = g_strdup_printf("(%s AND %s != '%s') ",string, criteria->field, criteria->value);
			}
			else if ( criteria->boolval  == LOOKUP_NOR )
			{
				string = g_strdup_printf("(%s OR %s != '%s') ",string, criteria->field, criteria->value);
			}
		}
		counter++;	
		tmp = g_slist_next(tmp);
	}
	g_print("persistant_lookup_criteria: %s\n",string);
	return string;
}


static void
cfdb_create_update_string(GtkWidget *cf, GtkWidget *changed)
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(changed != NULL);
	g_return_if_fail(GTK_IS_WIDGET(changed));
	g_return_if_fail(IS_CFDB(cf));
	
	if ( GTK_IS_EDITABLE(changed) )
	{
		gchar* field = g_hash_table_lookup(CFDB(cf)->widget_field_hash, changed);
		if (CFDB(cf)->number_fields_update == 0)
		{
			/* if this is the first or only field, don't put a comma in the update sql string */
		  CFDB(cf)->update_string = g_strdup_printf(" set %s = '%s' ",field, gtk_editable_get_chars(GTK_EDITABLE(changed),0,-1) );
		  CFDB(cf)->number_fields_update++;
		}
		else		
		{	
			/* not the first or only so put a comma in the update sql string */
			CFDB(cf)->update_string = g_strdup_printf("%s %s", CFDB(cf)->update_string, g_strdup_printf(", set %s = '%s' ",field, gtk_editable_get_chars(GTK_EDITABLE(changed),0,-1)));
			CFDB(cf)->number_fields_update++;
		}
		g_print("In create_update_string: is editable end: number_fields_update: %d, string: %s\n", CFDB(cf)->number_fields_update, CFDB(cf)->update_string);
	}
	else
	{
		g_warning("create_sql: don't know how to handle widget type!\n");	
	}
}


/* called when a registered widget changes (text/values) */
static void
cfdb_widget_changed(GtkWidget *changed, GtkWidget *cf)
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	CFDB(cf)->field_changed = TRUE;
}

static void
cfdb_widget_lost_focus(GtkWidget *changed, GdkEventFocus *event, gpointer cf)
{
 	g_return_if_fail(changed != NULL);
 	g_return_if_fail(cf != NULL);
 	g_return_if_fail(GTK_IS_WIDGET(changed));
	g_return_if_fail(IS_CFDB(cf));
	
	/* if we are in normal mode and there has been a widget change, Partially make update_string
		ready for when we reload or move */
  if ( CFDB(cf)->current_mode == MODE_NORMAL )
  {
//		g_print("In cfdb_widget_lost_focus: start\n");
		if ( CFDB(cf)->field_changed == TRUE )
		{	
//			g_print("In cfdb_widget_lost_focus: field_changed is TRUE\n");
			cfdb_create_update_string(GTK_WIDGET(cf), changed);
			CFDB(cf)->field_changed = FALSE;
		}
//		g_print("In cfdb_widget_lost_focus: end\n");
	}
}

static void
cfdb_widgets_activate(GtkWidget *widget, GtkWidget *cf)
{
 	g_return_if_fail(widget != NULL);
 	g_return_if_fail(cf != NULL);
 	g_return_if_fail(GTK_IS_WIDGET(widget));
	g_return_if_fail(IS_CFDB(cf));

	
 	if ( CFDB(cf)->current_mode == MODE_INSERT )
	{
		void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_insert_string;
		void (*hash_free)(gpointer, gpointer, gpointer) =  cfdb_hash_free;
		CFDB(cf)->number_fields_insert = 0;
		g_free(CFDB(cf)->insert_string1);
		g_free(CFDB(cf)->insert_string2);
		CFDB(cf)->insert_string1 = NULL;
		CFDB(cf)->insert_string2 = NULL;		

//     g_print("CFDB(user_data)->number_fields_insert: %d\n",CFDB(cf)->number_fields_insert);
//     g_print("IS1: %s IS2 %s \n",CFDB(cf)->insert_string1, CFDB(cf)->insert_string2);
		/* build the initial insert strings */
		g_hash_table_foreach(CFDB(cf)->field_widget_hash, hash_func, CFDB(cf));
		
		/* finish them off */
		if ( CFDB(cf)->insert_string1 != NULL && CFDB(cf)->insert_string2 != NULL )
		{

			CFDB(cf)->insert_string1 = g_strdup_printf("INSERT INTO %s (%s) values (%s)",CFDB(cf)->table,CFDB(cf)->insert_string1, CFDB(cf)->insert_string2);
			g_free(CFDB(cf)->insert_string2);
			CFDB(cf)->insert_string2 = NULL;
			g_hash_table_foreach(CFDB(cf)->insert_field_value_hash,hash_free,CFDB(cf)->insert_field_value_hash);
			g_hash_table_destroy(CFDB(cf)->insert_field_value_hash);
			CFDB(cf)->insert_field_value_hash = NULL;
			CFDB(cf)->insert_field_value_hash  = g_hash_table_new(g_str_hash, g_str_equal);
		
			/* do a reload which will actually do the insert */
			cfdb_do_reload(cf, RELOAD_ALL);
		}
		else
		{
			CFDB(cf)->current_mode = MODE_NORMAL;
			cfdb_do_reload(cf, RELOAD_CONTINUE);
		}
  }
}

static void
cfdb_hash_field_lookup(gpointer key, gpointer value, gpointer user_data)
{
	/* Changed to get rid of gchar* */
	if ( IS_CFDB(user_data) )
	{
		if(CFDB(user_data)->sql_string == NULL )
		{
			CFDB(user_data)->sql_string = g_strdup_printf("%s",(gchar*)key);
		}
		else
		{
			CFDB(user_data)->sql_string = g_strdup_printf("%s, %s",CFDB(user_data)->sql_string, (gchar*)key);
		}
	}
	else
	{
		g_print("Error user_data not cfdb\n");
	}
}

static void
cfdb_hash_insert_string(gpointer key, gpointer value, gpointer user_data)
{
	g_return_if_fail(key != NULL);
	g_return_if_fail(value != NULL);
	g_return_if_fail(user_data != NULL);
	g_return_if_fail(GTK_IS_WIDGET(value));
	g_return_if_fail(IS_CFDB(user_data));

//g_print("cfdb_hash_insert_string: key: %s  number_fields_insert: %d\n",key, CFDB(user_data)->number_fields_insert);
	if ( IS_CFDB(user_data) )
	{
		if ( GTK_IS_EDITABLE(value) )
		{
			/* we only want to do inserts on fields that have values */
			if ( strlen(gtk_editable_get_chars(GTK_EDITABLE(value),0,-1)) > 0 )
			{
				/* different the first time around ( no comma ) */
				if ( CFDB(user_data)->number_fields_insert == 0 )
				{
						CFDB(user_data)->insert_string1 = g_strdup_printf("%s",(gchar*) key);
						CFDB(user_data)->insert_string2 = g_strdup_printf("'%s'",g_strdelimit(gtk_editable_get_chars(GTK_EDITABLE(value),0,-1),'\'',"\'"));
				}
				else
				{
						CFDB(user_data)->insert_string1 = g_strdup_printf("%s,%s",CFDB(user_data)->insert_string1, (gchar*) key);
						CFDB(user_data)->insert_string2 = g_strdup_printf("%s,'%s'",CFDB(user_data)->insert_string2,g_strdelimit(gtk_editable_get_chars(GTK_EDITABLE(value),0,-1),"'","\'"));
				}
				/* insert the field and */
				g_hash_table_insert(CFDB(user_data)->insert_field_value_hash, (gchar*) key, gtk_editable_get_chars(GTK_EDITABLE(value),0,-1));
				CFDB(user_data)->number_fields_insert++;
				g_print("cfdb_hash_insert_string: IS1: %s IS2: %s\n",CFDB(user_data)->insert_string1, CFDB(user_data)->insert_string2);
			}
		}
		else
		{
			g_warning("cfdb_hash_insert_string: don't know how to handle widget\n");
		}
	}
	else
	{
		g_warning("Warning user_data not type cfdb\n");
	}
}

static void
cfdb_block_all_widgets(GtkWidget *cf)
{
 	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_widget_block;
	g_hash_table_foreach(CFDB(cf)->widget_field_hash, hash_func, CFDB(cf));
}

static void
cfdb_hash_widget_block(gpointer key, gpointer value, gpointer user_data)
{
	g_return_if_fail(key != NULL);
	g_return_if_fail(user_data != NULL);
	g_return_if_fail(GTK_IS_OBJECT(key));
	g_return_if_fail(IS_CFDB(user_data));

	/* registered child cfdb widgets don't need to be blocked */
	if ( !IS_CFDB(key) )
		gtk_signal_handler_block_by_data(GTK_OBJECT(key), user_data);
}

static void
cfdb_unblock_all_widgets(GtkWidget *cf)
{
 	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_widget_unblock;
	g_hash_table_foreach(CFDB(cf)->widget_field_hash, hash_func, CFDB(cf));
}

static void
cfdb_hash_widget_unblock(gpointer key, gpointer value, gpointer user_data)
{
	g_return_if_fail(key != NULL);
	g_return_if_fail(user_data != NULL);
	g_return_if_fail(GTK_IS_OBJECT(key));
	g_return_if_fail(IS_CFDB(user_data));
	
	/* registered child cfdb widgets don't need to be blocked */	
	if ( !IS_CFDB(key) )
		gtk_signal_handler_unblock_by_data(GTK_OBJECT(key), user_data);
}

void
cfdb_set_insert_mode(GtkWidget *widget, GtkWidget *cf)
{
	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_clear_all_widgets;
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	CFDB(cf)->current_mode = MODE_INSERT;
	gtk_label_set_text(GTK_LABEL(CFDB(cf)->labelRecordCounter), "NEW");
	g_hash_table_foreach(CFDB(cf)->widget_field_hash, hash_func, CFDB(cf));

}

static void
cfdb_hash_clear_all_widgets(gpointer key, gpointer value, gpointer user_data)
{
	
	g_return_if_fail(key != NULL);
	g_return_if_fail(user_data != NULL);
	g_return_if_fail(GTK_IS_OBJECT(key));
	g_return_if_fail(IS_CFDB(user_data));
	
	if ( GTK_IS_EDITABLE(key) )
	{
		gtk_editable_delete_text(GTK_EDITABLE(key),0,-1);
	}
}

static void
cfdb_hash_free(gpointer key, gpointer value, gpointer user_data)
{
 	g_hash_table_remove((GHashTable*) user_data, key);
}

void
cfdb_add_lookup_criteria( GtkWidget *cf, gint boolval, gchar* field, gchar* value)
{
	CfDbSearch *criteria = NULL;

	g_return_if_fail(cf != NULL);
	g_return_if_fail(field != NULL);
	g_return_if_fail(value != NULL);
	g_return_if_fail(IS_CFDB(cf));

	criteria = g_malloc(sizeof(CfDbSearch));	
	
	if ( boolval != LOOKUP_AND && boolval != LOOKUP_OR && boolval != LOOKUP_NAND && boolval != LOOKUP_NOR)
		criteria->boolval = LOOKUP_AND; /* default to AND lookup */
	else
		criteria->boolval = boolval;
		
	criteria->field = field;
	criteria->value = value;
	
	CFDB(cf)->lookup_criteria = g_slist_append(CFDB(cf)->lookup_criteria, criteria);
}

void
cfdb_add_persistant_lookup_criteria( GtkWidget *cf, gint boolval, gchar* field, gchar* value)
{
	CfDbSearch *criteria = NULL;

	g_return_if_fail(cf != NULL);
	g_return_if_fail(field != NULL);
	g_return_if_fail(value != NULL);
	g_return_if_fail(IS_CFDB(cf));

	criteria = g_malloc(sizeof(CfDbSearch));	
	
	if ( boolval != LOOKUP_AND && boolval != LOOKUP_OR  && boolval != LOOKUP_NAND && boolval != LOOKUP_NOR)
		criteria->boolval = LOOKUP_AND; /* default to AND lookup */
	else
		criteria->boolval = boolval;
		
	criteria->field = field;
	criteria->value = value;
	
	CFDB(cf)->persistant_lookup_criteria = g_slist_append(CFDB(cf)->persistant_lookup_criteria, criteria);
}


void
cfdb_remove_all_lookup_criteria( GtkWidget *cf )
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	g_slist_foreach(CFDB(cf)->lookup_criteria,cfdb_slist_free,NULL);
	g_slist_free(CFDB(cf)->lookup_criteria);
	CFDB(cf)->lookup_criteria = NULL;
}

void
cfdb_remove_all_persistant_lookup_criteria( GtkWidget *cf )
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
	g_slist_foreach(CFDB(cf)->persistant_lookup_criteria,cfdb_slist_free,NULL);
	g_slist_free(CFDB(cf)->persistant_lookup_criteria);
	CFDB(cf)->persistant_lookup_criteria = NULL;
}

static void
cfdb_slist_free( gpointer data, gpointer user_data )
{
	g_free(data);
}

static void
cfdb_slist_count( gpointer data, gpointer user_data )
{
	GPOINTER_TO_INT(user_data)++;
}

void
cfdb_delete_current_record( GtkWidget *cf )
{
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));

	/*delete the current record based on the current primary key value */
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, g_strdup_printf("DELETE FROM %s where %s = '%s'",CFDB(cf)->table, CFDB(cf)->primarykey, CFDB(cf)->primarykey_value));
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("DELETE command failed\n");
		g_warning("Error Message: %s\n",PQresultErrorMessage(CFDB(cf)->res));
		g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(CFDB(cf)->res)));
		PQclear(CFDB(cf)->res);
	}
	PQclear(CFDB(cf)->res);
		
	/* reload to the same position (will be different record but close to original pre-delete position) */
	cfdb_do_reload(cf, RELOAD_CONTINUE);

	/* for some reason it doesn't reload continue the first time */
  cfdb_do_reload(cf, RELOAD_CONTINUE);

	/* code to attempt to create a backup of deleted data into another table.  Will implement at a later date ashc - 18062000*/
//	cf->res = PQexec(cf->conn, g_strdup_printf("SELECT tablename from pg_tables where tablename = '%s_deleted'",CFDB(cf)->table);
//	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_TUPLES_OK)
//	{
//		g_warning("SELECT command failed\n");
//		g_warning("Error Message: %s\n",PQresultErrorMessage(cf->res));
//		g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(cf->res)));
//		PQclear(CFDB(cf)->res);
//	}
//
//	if ( PQntuples(CFDB(cf)->res == 0)
//	{
//	
//		cf->res = PQexec(CFDB(cf)->conn, g_strdup_printf("SELECT *, now() AS date_deleted INTO %s_deleted);
//		if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
//		{
//			g_warning("UPDATE command failed: %s\n", CFDB(cf)->update_string);
//			g_warning("Error Message: %s\n",PQresultErrorMessage(CFDB(cf)->res));
//			g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(CFDB(cf)->res)));
//			PQclear(CFDB(cf)->res);
//		}
//		PQclear(CFDB(cf)->res);
//	}
}

/* DANGEROUS FUNCTION, COULD DELETE ENTIRE TABLE IF ASKED TO.  PLEASE ASK USER FOR CONFIRMATION!!!! */
void
cfdb_delete_current_lookup( GtkWidget *cf )
{
	gchar* delete_string = NULL;
	
	g_return_if_fail(cf != NULL);
	g_return_if_fail(IS_CFDB(cf));
	
  delete_string = g_strdup_printf("DELETE FROM %s",CFDB(cf)->table);
	if ( CFDB(cf)->lookup_criteria != NULL )
	{
		delete_string =g_strdup_printf(" WHERE %s",	cfdb_create_lookup_criteria(cf));
	}
	
	CFDB(cf)->res = PQexec(CFDB(cf)->conn, delete_string);
	if (!CFDB(cf)->res || PQresultStatus(CFDB(cf)->res) != PGRES_COMMAND_OK)
	{
		g_warning("DELETE command failed\n");
		g_warning("Error Message: %s\n",PQresultErrorMessage(CFDB(cf)->res));
		g_warning("Result Status: %s\n",PQresStatus(PQresultStatus(CFDB(cf)->res)));
		PQclear(CFDB(cf)->res);
	}
	PQclear(CFDB(cf)->res);
	
	/* reload to the same position (will be different record but close to original pre-delete position) */
	cfdb_do_reload(cf, RELOAD_ALL);

}

int
cfdb_get_number_of_records( GtkWidget *cf )
{
	g_return_val_if_fail(cf != NULL, -1);
	g_return_val_if_fail(IS_CFDB(cf), -1);

	return CFDB(cf)->totalCount;
}

#warning "cfdb_get_number_of_registered_fields(GtkWidget *cf) doesn't work.\nHash function doesn't pass the list back (??!!)"
int
cfdb_get_number_of_registered_fields(GtkWidget *cf)
{
	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_count_registered_fields;
	gint numfields = 0;
	g_return_val_if_fail(cf != NULL, -1);
	g_return_val_if_fail(IS_CFDB(cf), -1);

	g_hash_table_foreach(CFDB(cf)->field_widget_hash, hash_func, GINT_TO_POINTER(numfields));
	
	return numfields;
}

static void
cfdb_hash_count_registered_fields(gpointer key, gpointer value, gpointer user_data)
{
	g_return_if_fail(key != NULL);
	g_return_if_fail(user_data != NULL);
	
	GPOINTER_TO_INT(user_data)++;
}

#warning "cfdb_get_registered_fields(GtkWidget *cf) doesn't work.\n Hash function doesn't pass the list back (??!!)"
gchar** 		
cfdb_get_registered_fields(GtkWidget *cf)
{
// 	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_get_registered_fields;
	gchar** field_list;
	guint i, n = 1;
	GSList *slist;
	GSList *flist = NULL;
	
	g_return_val_if_fail(cf != NULL,NULL);
	g_return_val_if_fail(IS_CFDB(cf),NULL);

	g_print("&& %d\n",cfdb_get_number_of_registered_fields(cf));
	g_hash_table_foreach(CFDB(cf)->field_widget_hash, cfdb_hash_get_registered_fields, (gpointer)flist);
		
	g_slist_foreach(flist,cfdb_slist_count,GINT_TO_POINTER(n));
	
	field_list = g_new(gchar*, cfdb_get_number_of_registered_fields(cf) + 1);
	i = n;
	field_list[i--] = NULL;
	
	for (slist = flist ; slist ; slist = g_slist_next(slist) )
	{
		field_list[i--] = slist->data;
	}
	
	/* free the list but not the elements (as they are now in field_list) */
	g_slist_free(flist);
	
	return field_list;
}


void
cfdb_hash_get_registered_fields(gpointer key, gpointer value, gpointer user_data)
{
	GSList *list = (GSList*) user_data;
	g_return_if_fail(key != NULL);
	
	list = g_slist_prepend(list,key);
	user_data = (gpointer)list;
}

#warning "cfdb_get_all_fields(GtkWidget *cf) doesn't work.\n Hash function doesn't pass the list back (??!!)"
gchar** 		
cfdb_get_all_fields(GtkWidget *cf)
{
 	void (*hash_func)(gpointer, gpointer, gpointer) =  cfdb_hash_get_registered_fields;
	gchar** field_list;
	guint i, n = 1;
	GSList *flist = NULL, *slist;
	
	g_return_val_if_fail(cf != NULL,NULL);
	g_return_val_if_fail(IS_CFDB(cf),NULL);

	g_hash_table_foreach(CFDB(cf)->field_widget_hash, hash_func, (gpointer)flist);
	
	g_slist_foreach(flist,cfdb_slist_count,GINT_TO_POINTER(n));
	
	field_list = g_new(gchar*, n);
	i = n;
	field_list[i--] = NULL;
	
	for (slist = flist ; slist ; slist = g_slist_next(slist) )
	{
		field_list[i--] = slist->data;
	}
	
	/* free the list but not the elements (as they are now in field_list) */
	g_slist_free(flist);
	
	return field_list;
}


//int
//cfdb_get_total_number_of_fields(GtkWidget *cf)
//{
//	PGconn *tmpconn = NULL;
//	PGresult *res = NULL;
//	
//	gint numfields = 0;
//	g_return_val_if_fail(cf != NULL, -1);
//	g_return_val_if_fail(IS_CFDB(cf), -1);
//
//	tmpconn = PQconnectdb(CFDB(cf)->conn_info);
//  if (PQstatus(tmpconn) == CONNECTION_BAD)
//	{
//		g_warning("Temporary Connection to database '%s' failed. Connection string: %s\n", PQdb(tmpconn), CFDB(cf)->conn_info);
//		g_warning("%s", PQerrorMessage(CFDB(cf)->conn));
//		PQfinish(tmpconn);
//		return NULL;
//	}
//	
//
//	
//	return numfields;
//}

