/***************************************************************************
                          cfdb.h  -  description
                             -------------------
    begin                : Sat Jul  1 09:18:20 EST 2000
    copyright            : (C) 2000 by Ashley Cambrell
    email                : bec_ash@bigpond.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __CFDB_H__
#define __CFDB_H__


#include <gdk/gdk.h>
#include <gtk/gtkvbox.h>
#include <libpq-fe.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define CFDB(obj)          GTK_CHECK_CAST (obj, cfdb_get_type (), CfDb)
#define CFDB_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, cfdb_get_type (), CfDbClass)
#define IS_CFDB(obj)       GTK_CHECK_TYPE (obj, cfdb_get_type ())


typedef struct _CfDb       CfDb;
typedef struct _CfDbClass  CfDbClass;
typedef struct _CfDbSearch CfDbSearch;

struct _CfDb
{
  GtkVBox vbox;  /* inherits from GtkVBox (might not be the best choice) */

	GtkWidget *parent_cfdb; /* parent cfdb  for one - to - many relationships */

  GtkWidget *labelRecordCounter; /* label to show record count */

  GtkWidget *buttonStart; /* button to go to start of records */

  GtkWidget *buttonPrev;  /* button to go to previous record */

  GtkWidget *buttonNext; /* button to go to next record */

  GtkWidget *buttonLast; /* button to go to end of records */

	int totalCount; /* total number of records */

	int currentCount; /* current record number */

	gchar *primarykey; /* name of the primarykey */
	
	gchar *primarykey_value; /* current value of the primarykey */
			
	PGconn* conn; /* the database connection.  Each instance must have
									it's own connection as mulitple cursors don't work
									with the same conn */
	
	PGresult *res;  /* current result from the database */
	
	gchar* conn_info;
	
	gchar* sql_string; /* current sql_string which is built from the field - widget mappings */
	
	gchar* update_string; /* update_string which is built from the field - value mappings when updates are needed */
	
	gchar* cursor_name; /* current cursor name which is unique for each begin commit block - automatic */
	
	gchar* user; /* usename - gotten from $USER environment variable */
	
	gchar* table; /* name of the table for this cfdb table - NB: at the momoment you can only have one table per
										cfdb.  select a.x, b.x from a, b where a.z = b.z does not work! */
	
	GHashTable *field_widget_hash;  /* field - widget map */
	
	GHashTable *widget_field_hash; /* widget - field map */
	
 	GHashTable *insert_field_value_hash; /* field - value hash so we can position the cursor to the correct record after an insert */

//	GSList *changed_field_values; /* list of fields and values that have changed */
	
//	GSList *widget_handler_ids; /* list of handler id's that are mapped so we can block signals when we change data */
	
	int number_fields_update;  /* counter of how many fields have changed since we moved or reloaded */
	
	gboolean field_changed; /* so we know whether a field changed when we lose focus */
	
	gboolean *being_reloaded;  /* internal identifier so that the "moved" signal isn't emitted too early */
	
	guint current_mode;
	
	gchar* insert_string1; /* 2 insert strings so we can more easily build a dynamic insert string */
	
	gchar* insert_string2; /* 2 insert strings so we can more easily build a dynamic insert string */
	
	int number_fields_insert;
	
	GSList *lookup_criteria; /* used to narrow or add to a search (dynamically create a where clause) */
	
	GSList *persistant_lookup_criteria; /* lookup criteria that is not killed by remove_lookup_criteria */
};

struct _CfDbClass
{	
	GtkVBoxClass parent_class;

  void (* moved) (CfDb *cf); /* moved signal */
};

struct _CfDbSearch
{
	gint boolval;
	gchar* field;
	gchar* value;
};

enum {
	LOOKUP_AND,
	LOOKUP_OR,
	LOOKUP_NAND,
	LOOKUP_NOR
};

guint       cfdb_get_type        			(void);
//GtkWidget*  cfdb_new             			(void);

/* cfdb_new takes conn_info: "host=xxx port=xxx dbname=xxx user=xxxx (if user is not specified then current $USER is used) password=xxx */
GtkWidget* 	cfdb_new (gchar *conn_info, gchar *primarykey, gchar *table);//////, GtkWidget *parent_cf);

/* set the primarykey. Is also done via cfdb_new */
void				cfdb_set_primarykey				(GtkWidget *cf, gchar* primarykey);

/* set the primarykey value (useful for relational things) */
void cfdb_set_primaryley_value(GtkWidget* cf, gchar* primarykey_value);

/* do a complete reload */
void 				cfdb_reload								(GtkWidget *cf);

/* do a reload and go back to the same position */
void 				cfdb_reload_continue			(GtkWidget *cf);

/* set the connection for the database */
void				cfdb_set_conn							(GtkWidget *cf, PGconn *conn);

/* add a field - widget map. This also sets a widget - field map */
void				cfdb_add_field_widget_map (GtkWidget *cf, gchar* field, GtkWidget* widget);
																																													
/* set the table name */																																													
void 				cfdb_set_table						(GtkWidget *cf, gchar* table);

/* put the widget into insert mode (clear contents of all widgets and ready everything for inserting of stuff */
void				cfdb_set_insert_mode			(GtkWidget *widget, GtkWidget *cf);

/* gets the number of fields that are mapped with the cfdb widget */
int					cfdb_get_number_of_registered_fields	(GtkWidget *cf);

/* returns a NULL terminated array of all the fields mapped with the cfdb widget - be sure to free all the pointers
 once you are done with them */
gchar** 		cfdb_get_registered_fields						(GtkWidget *cf);
/* add lookup criteria to narrow or expand current lookup  - boolval = LOOKUP_AND for AND, bollval = LOOKUP_OR for OR */
void 				cfdb_add_lookup_criteria(GtkWidget *cf, gint boolval, gchar* field, gchar* value);

/* remove all the current lookup criteria */
void 				cfdb_remove_all_lookup_criteria(GtkWidget *cf);

/* add lookup criteria that is usually unchanging and is exact ( uses = not ~* ) and overrides standard lookcriteria
	This is ideal if you are writing a multi - user application where you anyone can see their own records (private &
	public) and everyone else's public records (but not private) */
void 				cfdb_add_persistant_lookup_criteria(GtkWidget *cf, gint boolval, gchar* field, gchar* value);

/* remove all the persistant lookup criteria */
void 				cfdb_remove_persistant_all_lookup_criteria(GtkWidget *cf);
/* get the number of records in the current lookup */
int 				cfdb_get_number_of_records(GtkWidget *cf);

/* deletes all records in the current lookup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 DANGEROUS FUNCTION, COULD DELETE ENTIRE TABLE IF ASKED TO (eg no lookup criteria).
	PLEASE ASK USER FOR CONFIRMATION!!!! */
void 				cfdb_delete_current_lookup(GtkWidget *cf);

/* delete the current record */
void 				cfdb_delete_current_record(GtkWidget *cf);

/* After cfdb_new and ALL widgets are registered, this starts the ball rolling */
void				cfdb_start						(GtkWidget *cf);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __CfDb_H__ */
